import json
import pathlib
import platform
import os
from typing import Any

import yaml
import streamlit as st
from pathvalidate import is_valid_filepath

from gwstreamlit import constants
from gwstreamlit.constants import KeyType
from gwstreamlit.models import InputFieldsBase, Tab, Button, BaseConfig, UserInterface


def codeify_string(input_string: str):
    new_string = input_string.replace(' ', '_').lower().replace("_/_", "_")
    return new_string


def __create_key(ui_item: any) -> str:
    """Create a key for the item, the key is used to store the value in the session state
    @param ui_item: The item to create the key for"""
    gw_streamlit = st.session_state["GWStreamlit"]
    if isinstance(ui_item, str):
        model_item = gw_streamlit.find_model_part(ui_item)
        return __create_key(model_item)
    if isinstance(ui_item, InputFieldsBase):
        return create_key_process(KeyType.INPUT, ui_item, ["Code", "Name", "Label"])
    if isinstance(ui_item, Tab):
        return create_key_process(KeyType.TAB, ui_item, ["Code", "Label"])
    if isinstance(ui_item, Button):
        return create_key_process(KeyType.BUTTON, ui_item, ["Code", "Label"])


def create_simple_key(key_type: KeyType, value: str):
    gw_streamlit = st.session_state["GWStreamlit"]
    application = gw_streamlit.application
    key = codeify_string(input_string=f"{key_type.value}_{application}_{value}")
    return key


def _create_storage_key(value: str):
    gw_streamlit = st.session_state["GWStreamlit"]
    application = gw_streamlit.application
    key = codeify_string(input_string=f"{KeyType.STORAGE.value}_{application}_{value}")
    return key


def create_key_process(key_type: KeyType, ui_item: any, keys: list):
    gw_streamlit = st.session_state["GWStreamlit"]
    application = gw_streamlit.application
    value = "undefined"
    for key in keys:
        if key in ui_item.model_fields.keys():
            value = getattr(ui_item, key)
            if value is not None:
                break
    key = codeify_string(input_string=f"{key_type.value}_{application}_{value}")
    return key


def read_yaml(yaml_file: str):
    with open(yaml_file, 'r') as file:
        yaml_data = yaml.safe_load(file)
        return yaml_data


def get_config_path(directory, file_name: str):
    """Returns the path to the configuration file depending on the operating system."""
    if file_name.endswith('.json'):
        config_filename = file_name
    else:
        config_filename = f"{file_name}.json"
    if platform.system() == 'Windows':
        # On Windows, it's typical to store config files in the AppData directory
        config_directory = os.path.join(os.getenv('APPDATA'), directory)
    elif platform.system() == 'Darwin':
        # On macOS, it's typical to store config files in the Application Support directory
        user_directory = os.path.expanduser('~/Library/Application Support/')
        config_directory = os.path.join(user_directory, "Field Framework", directory)
    else:
        raise OSError("Unsupported operating system")

    if not os.path.exists(config_directory):
        os.makedirs(config_directory)  # Create the directory if it does not exist

    return os.path.join(config_directory, config_filename)


def disabled(item: BaseConfig) -> bool:
    """Check if the item is enabled"""
    disabled_value = fetch_boolean(getattr(item, "immutable", False))
    if item.Enabled is not None:
        ...
        # if st.session_state.get(create_key(item.get("enabled")), None) is None:
        #     disabled_value = True
        # else:
        #     disabled_value = False
    return disabled_value


def fetch_boolean(value):
    if type(value) is bool:
        return value
    if type(value) is str:
        if value.lower() == "true":
            return True
        elif value.lower() == "false":
            return False


class InputField:
    pass


def build_label(item: BaseConfig):
    if item.Required:
        return f"{item.Label} *"
    else:
        return item.Label





def to_list(item_value) -> list:
    """Convert the item value to a list if it is not already a list
    :arg item_value: Value to be converted to a list
    """
    if type(item_value) is list:
        return item_value
    else:
        return [item_value]


def updated_edited_rows(df, edited_item):
    """Update the edited rows in the dataframe"""
    for key, value in edited_item.items():
        for item_key, item_value in value.items():
            df.loc[key, item_key] = item_value
    return df


def update_data_editor(*, key: str, replace_values: dict):
    update_dataframe(key=key, update_rows=replace_values)


def update_dataframe(key: str, update_rows: dict):
    """Updates a dataframe, the dataframe should have been built and stored in self.data_frame
    The original rows are removed from the dataframe and the new rows are added. As there can be a missmatch in the
    index sizes this is the easiest way to perform the update of the dataframe. This function is only used to update
    the dataframe in the session_state from a saved configuration, otherwise the changes will be available in the
    streamlit data_edit component.
    :param key: The key used to store the dataframe in the session_state
    :param update_rows: The rows to be updated in the dataframe"""
    df_key = f"{key}_df"
    if st.session_state.get(df_key, None) is None:
        return

    st.session_state.get(df_key)

    original_index = st.session_state.get(df_key).index
    original_index_list = list(range(original_index.start, original_index.stop))
    st.session_state.get(df_key).drop(original_index_list, inplace=True)
    for item in update_rows:
        st.session_state.get(df_key).loc[len(st.session_state.get(df_key))] = item
    st.session_state.get(df_key).reset_index(drop=True, inplace=True)


def _load_config(file_name):
    gws = st.session_state["GWStreamlit"]
    if file_name is None:
        return
    if pathlib.Path(file_name).name == file_name:
        directory = codeify_string(input_string=gws.application)
        config_path = get_config_path(directory, file_name)
    else:
        config_path = file_name
    try:
        with open(config_path, 'r') as file:
            config = json.load(file)
            for key, value in config.items():
                if str(key).startswith("input_"):
                    if type(value) is list:
                        update_data_editor(key=key, replace_values=value)
                    else:
                        st.session_state[key] = value
    except FileNotFoundError:
        return


def _completed_required_fields() -> bool:
    """Show the required fields that are not filled
    @return: True if all required fields are filled, False if there are required fields that are not filled"""
    required_list = []
    model = st.session_state["GWStreamlit"].model
    for input_field in [model_input for model_input in model.Inputs if model_input.Required]:
        if input_field.Required and st.session_state.get(input_field.Key) is None:
            required_list.append(input_field.Label)
    if len(required_list) > 0:
        st.error(f"The following required fields are not filled: {', '.join(required_list)}")
        return False
    return True


def _write_string(location, file_name, content, **kwargs):
    """Write a string to a file, there are multiple checks to ensure the file is written correctly
    If the path is invalid it will return an error message, if the contents to write are None it will return a message
    if the location is valid but does not exist it will be created.

    Args: location: str: The location to write the file
          file_name: str: The name of the file
          content: str: The content to write to the file
          **kwargs: dict: Additional arguments to specify the package and extension
    """
    gw_streamlit = st.session_state["GWStreamlit"]
    for key, value in kwargs.items():
        if key == "package":
            package_parts = value.split(".")
            for package_part in package_parts:
                location = os.path.join(location, package_part)
        if key == "extension":
            file_name = f"{file_name}.{value}"

    if content is None:
        gw_streamlit.fetch_tab("Output").write(f"File content for: {location}/{file_name} is None")
        return
    if not is_valid_filepath(location, platform="auto"):
        gw_streamlit.fetch_tab("Output").error("Source Location is an invalid path")
        return
    is_exist = os.path.exists(location)
    if not is_exist:
        os.makedirs(location)
        gw_streamlit.fetch_tab("Output").write(f"Directory created: {location}")
    with open(f"{location}/{file_name}", "w") as file:
        file.write(content)
    gw_streamlit.fetch_tab("Output").write(f"File created: {location}/{file_name}")


def _fetch_key(ui_item: Any):
    """Fetch the key for the item, if the item is a string it will find the item in the model and return the key
    otherwise the key will be returned from the item
    @param ui_item: Model part or part identifier to fetch the key for"""
    if isinstance(ui_item, str):
        item_code = codeify_string(ui_item)
        gw_streamlit = st.session_state["GWStreamlit"]
        model_item = gw_streamlit.find_model_part(item_code)
        return _fetch_key(model_item)
    else:
        return ui_item.Key




def build_model(yaml_file):
    """Build the model from the yaml file, and update to add the key and code if missing to the model
    @param yaml_file: The yaml file to build the model from"""
    model = UserInterface.model_validate(yaml_file)
    for input_model in model.Inputs:
        if input_model.Code is None:
            model_code = codeify_string(input_model.Label)
            setattr(input_model, "Code", model_code)
        model_key = __create_key(input_model)
        setattr(input_model, "Key", model_key)

    for button_model in model.Buttons:
        if button_model.Code is None:
            model_code = codeify_string(button_model.Label)
            setattr(button_model, "Code", model_code)
        model_key = __create_key(button_model)
        setattr(button_model, "Key", model_key)
    return model


def find_yaml_ui(yaml_file_name: str):
    templates = st.session_state.get("templates", list_files(constants.YAML_UI_LOCATION, ['.yaml', '.yml']))
    yaml_object_list = [template for template in templates if template["name"] == yaml_file_name]

    if len(yaml_object_list) == 0:
        st.session_state["template_selection"] = None
        return
    yaml_object = yaml_object_list[0]
    st.session_state["template_selection"] = yaml_object["name"]
    return yaml_object


def find_yaml_other(yaml_file_name: str):
    yaml_object = load_yaml(yaml_file_name)
    st.session_state["template_selection"] = yaml_object["name"]
    return yaml_object


def list_files(directory_path, file_types: list):
    found_files = []
    for root, dirs, files in os.walk(directory_path):
        for file in files:
            file_extension = pathlib.Path(file).suffix
            if file_extension in file_types:
                found_files.append(load_yaml(os.path.join(str(root), file)))

    return found_files


def load_yaml(file_path: str):
    with open(file_path, 'r') as file:
        return yaml.safe_load(file)


def _create_saved_state():
    """Creates a saved state for the application"""
    saved_dict = {}
    for key, value in [item for item in st.session_state.items()]:
        if key.startswith("input_"):
            if key.endswith("_df"):
                continue
            if len([item for item in st.session_state.keys() if item == f"{key}_df"]) > 0:
                df = st.session_state[f"{key}_df"]
                for del_index in to_list(st.session_state[key]['deleted_rows']):
                    df.drop(del_index, inplace=True)
                for added_item in to_list(st.session_state[key]['added_rows']):
                    if added_item:
                        df.loc[len(df)] = added_item
                for edited_item in to_list(st.session_state[key]['edited_rows']):
                    if edited_item:
                        updated_edited_rows(df, edited_item)
                df.reset_index(drop=True, inplace=True)
                saved_dict[key] = df.to_dict("records")
            else:
                saved_dict[key] = value

        if key.startswith("storage_"):
            saved_dict[key] = value
    return saved_dict


def _fetch_configs(application_name: str):
    """List of saved configurations for the application"""
    file_list = []
    directory = codeify_string(application_name)
    config_path = get_config_path(directory, 'temp.json')
    for root, dirs, files in os.walk(os.path.dirname(config_path)):
        for file in files:
            if file.endswith('.json'):
                file_list.append(file)
    return file_list


def _save_config(application_name: str, file_name, config_data):
    """Saves the given configuration data to a JSON file."""
    if file_name is None:
        return
    directory = codeify_string(application_name)
    config_path = get_config_path(directory, file_name)
    with open(config_path, 'w') as file:
        json.dump(config_data, file, indent=4)


def _fetch_tab(item: Any):
    if isinstance(item, str):
        tab = st.session_state["GWStreamlit"].tab_dict.get(item)
    else:
        tab_name = item.Tab
        if tab_name is None:
            tab_name = "Main"
        tab = st.session_state["GWStreamlit"].tab_dict.get(tab_name)
    return tab




from gwstreamlit.models import BaseConfig


def construct_function(function_name):
    if function_name is None:
        return None
    function_module = function_name.split(":")[0]
    function_function = function_name.split(":")[1]
    defined_function = getattr(__import__(function_module, globals(), locals(), [function_function]), function_function)
    return defined_function


def option_function(item: BaseConfig):
    if len(item.InputOptions) == 1 and item.InputOptions[0].Function is not None:
        function_name = item.InputOptions[0].Function
        defined_function = construct_function(function_name)
        return defined_function
    else:
        return None


def get_default_rows(gws):
    default_rows_dict = {}
    for item in [table_inputs for table_inputs in gws.yaml_file.get("inputs", []) if
                 table_inputs.get("type") == "table"]:
        default_rows = item.get("default_rows", dict())
        default_rows_dict[item.get("label")] = default_rows
    return default_rows_dict



import streamlit as st
import gwstreamlit.core as gws
from jinja2 import FileSystemLoader, Environment, select_autoescape, TemplateNotFound


def _process_template_by_name(template_name, input_dict: dict):
    gw_streamlit = st.session_state["GWStreamlit"]
    env = Environment(
            loader=FileSystemLoader("resources/templates"),
            autoescape=select_autoescape(),
            trim_blocks=True,
    )
    template_result = None
    try:
        template = env.get_template(template_name)
        template_result = template.render(input_dict)
    except TemplateNotFound:
        gw_streamlit.fetch_tab("Output").warning(f"Template - {template_name} was not found")

    return template_result



from typing import Any

from pydantic import BaseModel, Field


class Option(BaseModel):
    Value: str = Field(alias="value", default=None)
    Function: str = Field(alias="function", default=None)


class BaseConfig(BaseModel):
    Code: str = Field(alias="code", default=None)
    Label: str = Field(alias="label", default=None)
    Enabled: str = Field(alias="enabled", default=None)
    OnClick: str = Field(alias="on_click", default=None)
    OnChange: str = Field(alias="on_change", default=None)
    Tab: str = Field(alias="tab", default="Main")
    Default: Any = Field(default=None, alias="default")
    DefaultFunction: str = Field(default=None, alias="default_function")
    Extension: str = Field(default=None, alias="extension")
    Immutable: bool = Field(default=False, alias="immutable")
    Key: str = Field(default=None, alias="key")
    InputOptions: list[Option] = Field(default=None, alias="options")


class Button(BaseConfig):
    Variant: str = Field(default="secondary", alias="variant")


class Tab(BaseConfig):
    pass


class InputFieldsBase(BaseConfig):
    Type: str = Field(default="text_input", alias="type")
    Required: bool = Field(default=False, alias="required")
    Min: int = Field(default=None, alias="min")
    Max: int = Field(default=None, alias="max")
    Order: str = Field(default=None, alias="order")
    Function: str = Field(alias="function", default=None)


class InputFields(InputFieldsBase):
    Image: str = Field(default=None, alias="image")
    Columns: list[InputFieldsBase] = Field(default=[], alias="columns")


class UserInterface(BaseConfig):
    Name: str = Field(alias="name")
    Description: str = Field(alias="description", default=None)
    Developer: str = Field(alias="developer")
    Concept: str = Field(alias="concept")
    Title: bool = Field(default=False, alias="title")
    Inputs: list[InputFields] = Field(default=[], alias="inputs")
    Tabs: list[Tab] = Field(default=[], alias="tabs")
    Buttons: list[Button] = Field(default=[], alias="buttons")



from pathlib import Path
from typing import Any

import streamlit as st

from gwstreamlit._create_ui import create_ui_title, create_ui_buttons, create_ui_tabs, create_ui_inputs, generate_image
from gwstreamlit._utils import get_default_rows
from gwstreamlit.process_templates import _process_template_by_name
from gwstreamlit.utils import find_yaml_ui, find_yaml_other, \
    build_model, _fetch_key, _fetch_configs, \
    _completed_required_fields, _create_saved_state, _save_config, _load_config, _write_string, _fetch_tab, \
    _create_storage_key


class GWStreamlit:

    def create_ui(self):
        """Builds the UI for the application"""
        create_ui_title(self)
        create_ui_buttons(self)
        if not self.model.Title:
            create_ui_tabs(self)
        create_ui_inputs(self)

    def fetch_tab(self, item: Any):
        if isinstance(item, str):
            tab = self.tab_dict.get(item)
        else:
            tab_name = item.Tab
            if tab_name is None:
                tab_name = "Main"
            tab = self.tab_dict.get(tab_name)
        return tab

    def fetch_model_input(self):
        ...

    def find_model_part(self, identifier: str):
        """Finds a model part by the identifier provided. The identifier can be the code or the
        label of the item. If the item is not found None is returned.
        @param identifier: str"""
        items = [item for item in self.model.Inputs if item.Code == identifier]
        if len(items) == 0:
            items = [item for item in self.model.Inputs if item.Label == identifier]
        if len(items) == 0:
            return None
        return items[0]

    def generate_image(self, item):
        generate_image(self, item)

    def __init__(self, application: str, yaml_file: dict):
        st.session_state["GWStreamlit"] = self
        self.application = application
        self.yaml_file = yaml_file
        self.model = build_model(self.yaml_file)
        self.keys = []
        self.input_values = {}
        self.tab_dict = {}
        self.default_rows = get_default_rows(self)


def initialize(application: str, yaml_file_name: str):
    if Path(yaml_file_name).name == yaml_file_name:
        yaml_file = find_yaml_ui(yaml_file_name)
    else:
        yaml_file = find_yaml_other(yaml_file_name)
    st.session_state["GWStreamlit"] = GWStreamlit(application, yaml_file)
    st.session_state["GWStreamlit"].create_ui()


def required_fields() -> bool:
    return _completed_required_fields()


def fetch_key(ui_item: Any) -> str:
    return _fetch_key(ui_item)


def fetch_configs(application_name: str) -> list:
    return _fetch_configs(application_name)


def create_saved_state():
    """Creates a saved state for the application"""
    return _create_saved_state()


def save_config(application_name: str, file_name, config_data):
    _save_config(application_name, file_name, config_data)


def load_config(file_name):
    _load_config(file_name)


def process_template_by_name(template_name, input_dict: dict):
    return _process_template_by_name(template_name, input_dict)


def write_string(location, file_name, content, **kwargs):
    _write_string(location, file_name, content, **kwargs)


def fetch_tab(item: Any):
    return _fetch_tab(item)


def create_storage_key(value: str):
    return _create_storage_key(value)



from typing import Optional

import streamlit as st
import pandas as pd

from gwstreamlit._utils import construct_function, option_function
from gwstreamlit.models import BaseConfig, InputFieldsBase
from gwstreamlit.utils import disabled, build_label, fetch_boolean


def create_ui_tabs(gws):
    """Creates a list of tabs to be used in the UI"""
    tab_dict = {item.Tab: None for item in gws.model.Inputs if item.Tab is not None}
    if len([item for item in gws.model.Inputs if item.Tab is None]) > 0:
        tab_dict["Main"] = None

    for tab in gws.model.Tabs:
        tab_dict[tab.Label] = None
    tab_dict["Output"] = None

    tabs = st.tabs(tab_dict.keys())
    tab_position = 0

    for tab in tab_dict.keys():
        tab_dict[tab] = tabs[tab_position]
        tab_position += 1
    gws.tab_dict = tab_dict


def create_ui_title(gws):
    """Creates a title for the UI page and a description if it exists in the YAML file.
    If the title exists in the yaml file no buttons are generated."""
    st.title(gws.model.Name)
    if gws.model.Description is not None:
        st.markdown(gws.model.Description)
    if gws.model.Concept is not None:
        st.write(f"Concept by: {gws.model.Concept}")
    if gws.model.Developer is not None:
        st.write(f"Developed by: {gws.model.Developer}")


def create_ui_buttons(gws):
    """Generates a set of buttons based on the YAML file provided"""
    columns = st.columns([1, 1, 1, 1, 1])
    column_index = 0
    for button in gws.model.Buttons:
        with columns[column_index]:
            on_click = construct_function(button.OnClick)
            st.button(f"{button.Label}", key=button.Key, on_click=on_click, type=button.Variant,
                      use_container_width=True)
        column_index += 1


def create_ui_inputs(gws):
    """Main processing for the inputs defined in the yaml file. Each type of input
    is handled separately. Each input has a key generated that corresponds to the label or code
    and the state name defined in the main module for the application"""

    if gws.model.Inputs is None:
        return
    for ui_item in gws.model.Inputs:
        build_input(gws, ui_item)


def build_input(gws, ui_item):
    """Main processing for the inputs defined in the yaml file. Each type of input
    is handled separately. Each input has a key generated that corresponds to the label or code
    and the state name defined in the main module for the application"""
    if ui_item.Key not in gws.input_values.keys():
        gws.input_values[ui_item.Key] = None

    if ui_item.Type == 'text_input':
        generate_text_input(gws, ui_item, None)

    if ui_item.Type == 'text_area':
        generate_text_input(gws, ui_item, None, True)

    if ui_item.Type == "selectbox":
        generate_selectbox(gws, ui_item, None)

    if ui_item.Type == "image":
        generate_image(gws, ui_item)

    if ui_item.Type == "checkbox":
        generate_checkbox(gws, ui_item, None)

    if ui_item.Type == "toggle":
        generate_checkbox(gws, ui_item, None, True)

    if ui_item.Type == "integer_input":
        generate_integer_input(gws, ui_item, None)

    if ui_item.Type == "file_upload":
        generate_file_upload(gws, ui_item, None)

    if ui_item.Type == "multiselect":
        generate_selectbox(gws, ui_item, None, True)

    if ui_item.Type == "table":
        generate_table(gws, ui_item)

    return ui_item.Key


def generate_text_input(gws, item, input_values, text_area=False):
    """text input processing. The default value is used if it exists in the YAML file. If the value is already
    in the input_values dictionary it is used instead of the default value. The on_change function is constructed
    based on the name provided in the YAML file. The input is disabled if the immutable flag is set in
    the YAML file."""
    default_value = item.Default
    if item.DefaultFunction:
        defined_function = construct_function(item.DefaultFunction)
        default_value = defined_function()
    if gws.input_values.get(item.Key, None) is not None:
        default_value = gws.input_values[item.Key]
    on_change = construct_function(item.OnChange)
    disabled_input = disabled(item)
    if text_area:
        gws.input_values[item.Key] = gws.fetch_tab(item).text_area(build_label(item), default_value, key=item.Key,
                                                                   on_change=on_change, disabled=disabled_input)
    else:
        gws.input_values[item.Key] = gws.fetch_tab(item).text_input(build_label(item), default_value,
                                                                    key=item.Key,
                                                                    on_change=on_change, disabled=disabled_input)


def generate_image(gws, item):
    if item.Image is not None:
        caption = f"{build_label(item)} - {item.Image}"
        gws.fetch_tab(item).image(item.Image, caption=caption)


def generate_checkbox(gws, item: BaseConfig, input_values, toggle=False):
    """Check Box Processing. The default value is used if it exists in the YAML file. If the value is already
    in the input_values dictionary it is used instead of the default value. The on_change function is constructed
    based on the name provided in the YAML file. The input is disabled if the immutable flag is set in the
    YAML file."""
    default_value = item.Default
    if gws.input_values.get(item.Key, None) is not None:
        default_value = gws.input_values[item.Key]
    on_change = construct_function(item.OnChange)
    disabled_field = fetch_boolean(item.Immutable)
    if not toggle:
        gws.input_values[item.Key] = gws.fetch_tab(item).checkbox(build_label(item), default_value, key=item.Key,
                                                                  on_change=on_change, disabled=disabled_field)
    else:
        gws.input_values[item.Key] = gws.fetch_tab(item).toggle(build_label(item), default_value, key=item.Key,
                                                                on_change=on_change, disabled=disabled_field)


def generate_file_upload(gws, item, input_values):
    """File Upload Processing. The extension is used to set the type of file that can be uploaded.
    The on_change function is constructed based on the name provided in the YAML file.
    The input is disabled if the immutable flag is set in the YAML file."""
    on_change = construct_function(item.OnChange)
    gws.input_values[item.Key] = gws.fetch_tab(item).file_uploader(build_label(item), type=item.Extension,
                                                                   accept_multiple_files=False, key=item.Key,
                                                                   on_change=on_change)


def generate_integer_input(gws, item: InputFieldsBase, input_values):
    """Integer Input Processing. The default value is used if it exists in the YAML file. If the value is already
    in the input_values dictionary it is used instead of the default value. The on_change function is constructed
    based on the name provided in the YAML file. The input is disabled if the immutable flag is set in the
    YAML file. The min and max values are used to set the range of the input."""
    if item.Min:
        min_value = item.Min
    else:
        min_value = 0

    if item.Max:
        max_value = item.Max
    else:
        max_value = 100

    if gws.input_values.get(item.Key, None):
        default_value = gws.input_values[item.Key]
    else:
        default_value = min_value
    on_change = construct_function(item.OnChange)
    disabled_field = fetch_boolean(item.Immutable)
    gws.input_values[item.Key] = gws.fetch_tab(item).number_input(build_label(item), value=default_value,
                                                                  key=item.Key, step=1, min_value=min_value,
                                                                  max_value=max_value, on_change=on_change,
                                                                  disabled=disabled_field)


def generate_selectbox(gws, item: BaseConfig, input_values, multiselect=False):
    """select box processing. The default value is used if it exists in the YAML file. If the value is already
    in the input_values dictionary it is used instead of the default value. The on_change function is constructed
    based on the name provided in the YAML file. The input is disabled if the immutable flag is set in the
    YAML file.
    If the options are defined as a function the function is called to get the options. If the default value is not
    in the options list the first option is used as the default value."""
    default_value = item.Default
    options_dict = options_list(gws, item)
    options = options_dict.get("options")
    default_value = options_dict.get("default_value")
    on_change = construct_function(item.OnChange)
    if options is not None and len(options) > 0 and default_value is not None:
        index_value = options.index(default_value)
    else:
        index_value = None
    with st.container():
        disabled_field = fetch_boolean(item.Immutable)
        if multiselect:
            gws.input_values[item.Key] = gws.fetch_tab(item).multiselect(build_label(item), options,
                                                                         default_value,
                                                                         key=item.Key, on_change=on_change,
                                                                         disabled=disabled_field)
        else:
            gws.input_values[item.Key] = gws.fetch_tab(item).selectbox(build_label(item), options, index_value,
                                                                       key=item.Key,
                                                                       on_change=on_change,
                                                                       disabled=disabled_field)


def generate_table(gws, item):
    """Main Processing for table generation. Each table is backed by a dataframe that is stored
    in self.data_frame."""
    default_rows = gws.default_rows.get(item.Label, dict())
    generate_dataframe(gws, item, default_rows)


def generate_dataframe(gws, item: InputFieldsBase, default_rows: dict):
    """Generate the Table Dataframe. The default rows are used to populate the dataframe if it is empty. The columns
    are generated based on the YAML file provided. The dataframe is stored in self.data_frame and can be updated"""
    if item.Columns is None:
        return

    if item.Columns[0].Function is not None:
        defined_function = construct_function(item.Columns[0].Function)
        columns = defined_function()
    else:
        columns = [entity_item.Label for entity_item in item.Columns]
    df_key = f"{item.Key}_df"
    if df_key in st.session_state.keys():
        df = st.session_state[df_key]
    else:
        df = pd.DataFrame(columns=columns, data=default_rows)
        if item.Order:
            df.sort_values(by=[item.Order], inplace=True, ignore_index=True)

    column_config = create_column_config(gws, item.Columns)
    gws.fetch_tab(item).markdown(f"**{item.Label}**")

    st.session_state[df_key] = df
    if item.Immutable:
        if default_rows:
            num_rows = len(default_rows)
        else:
            num_rows = 0
        gws.fetch_tab(item).data_editor(st.session_state[df_key], num_rows=num_rows,
                                        column_config=column_config, use_container_width=True, key=item.Key)
    else:
        gws.fetch_tab(item).data_editor(st.session_state[df_key], num_rows="dynamic", column_config=column_config,
                                        use_container_width=True, key=item.Key)


def create_column_config(gws, columns: list[InputFieldsBase]):
    """Define the column configuration for the data editor, multiselect is not supported in the data editor and
    list column is used"""
    column_config = {}
    for column_item in columns:
        column = column_item.Label
        if column is None:
            column = column_item
        if column_item.InputOptions:
            options_dict = options_list(gws, column_item)
            options = options_dict.get("options")
            default_value = options_dict.get("default_value")
            column_config[column] = st.column_config.SelectboxColumn(options=options, default=default_value)
        if column_item.Type == "checkbox":
            column_config[column] = st.column_config.CheckboxColumn(default=False)
        if column_item.Type == "integer_input":
            min_value = column_item.get('min', -1)
            max_value = column_item.get('max', 100)
            column_config[column] = st.column_config.NumberColumn(max_value=max_value, min_value=min_value, step=1,
                                                                  default=-1)
        if column_item.Type == "multiselect":
            column_config[column] = st.column_config.ListColumn()
    return column_config


def options_list(gws, item):
    defined_option_function = option_function(item)
    default_value: Optional[str] = None
    if defined_option_function is None:
        options = [option.Value for option in item.InputOptions]
        if item.Default not in options:
            default_value = item.InputOptions[0].Value
        if gws.input_values.get(item.Key, None) is not None:
            default_value = gws.input_values[item.Key]
    else:
        options = defined_option_function()
    return {"options": options, "default_value": default_value}



from enum import Enum


class KeyType(Enum):
    INPUT = "input"
    BUTTON = "button"
    TAB = "tab"
    STORAGE = "storage"


DEFAULT_BUTTONS = [
    {'label': 'Submit', 'on_click': 'gwstreamlit.utils:button_submit', 'type': 'submit', 'variant': 'primary'},
    {'label': 'Cancel', 'on_click': 'gwstreamlit.utils:button_cancel', 'type': 'cancel', 'variant': 'secondary'}
]

YAML_UI_LOCATION = "./resources/yaml_ui"
